# OBJ - timestamped JSON objects
#
# obj/__init__.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 08-02-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    object class to save/load json files.
    if a pathname is not provided a timestamp filename is provided, for example:
    the cfg.workdir variable in this module is prepended to a filename.

"""

__version__ = "10"

## imports

import logging
import os
import string
import time
import threading

_starttime = time.time()

## exceptions

class ESET(Exception):
    """ the provided string is not JSON. """
    pass

class EJSON(Exception):
    """ the provided string is not JSON. """
    pass

class ELOADER(Exception):
    """ loader is missing in event. """
    pass

class ENOJSON(Exception):
    """ the provided string is not JSON. """
    pass
    
class EBORDER(Exception):
    """ attemped to write outside of the workdir. """
    pass

class EFILENAME(Exception):
    """ provided string is not a filename. """
    pass

class ENOTSET(Exception):
    """ variable is not set. """
    pass

## helpers

def locked(func, *args, **kwargs):

    """ locked decorator, use to lock decorator function. """
    import _thread
    lock = _thread.allocate_lock()

    def lockedfunc(*args, **kwargs):
        """ lockded function to wrap a function in. """
        lock.acquire()
        res = None
        try:
            res = func(*args, **kwargs)
        finally:
            try:
                lock.release()
            except:
                pass
        return res
    return lockedfunc

## classes

class Object(dict):

    """ save/load object's state to timestamped json string file. """

    def __getattribute__(self, name):
        """ get attribute and, if that fails, check item access.  """
        if name == "__typed__":
            return self.__class__.__name__
        try:
            return super().__getattribute__(name)
        except AttributeError:
            try:
                return self[name]
            except KeyError:
                raise AttributeError(name)

    def __getattr__(self, name):
        """ get missing attribute by name. """
        if name == "_batch":
            self._batch = False
        if name == "_cbs":
            self._cbs = []
        if name == "_funcs":
            self._funcs = []
        if name == "_result":
            self._result = []
        elif name == "_parsed":
            self._parsed = Parsed(slice(self, ["cc", "txt"]))
        elif name == "_loader":
            self._loader = None
        elif name == "_container":
            self._container = Object()
        elif name == "_ready":
            self._ready = threading.Event()
        elif name == "_thrs":
            self._thrs = []
        if name == "id":
            self.id = repr(self)
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)
            
    def __repr__(self):
        """ mimic a normal __repr__. """
        return '<%s.%s at %s>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self))
        )

    def __setattr__(self, name, value):
        """ implement dotted dict access. """
        return self.__setitem__(name, value)

    def __str__(self):
        """ return a pretyified json string. """
        return self.nice()

    def add(self, txt):
        """ add text to result. """
        self._result.append(txt)

    def cache(self, path: str) -> 'Object':
        """ load this object from cache or read it from disk. """
        if path in cache:
            return cache[path]
        return self.load(path)

    def clear(self) -> None:
        """ clear the ready flag. """
        self._ready.clear()

    def dispatch(self):
        """ dispatch the object to the functions registered in the _funcs member. """
        for cb in self._cbs:
            cb(event)
        for func in self._funcs:
            func(self)

    def display(self, obj=None, keys=[], txt=""):
        """ display an object. """
        res = ""
        if not obj:
            obj = self
        if not keys:
            keys = self._parsed.args
        for key in keys:
            if key.startswith("_"):
                continue
            val = getattr(obj, key, None)
            if val:
                res += str(val).strip() + " "
        if txt and res:
            res = "%s %s" % (txt.strip(), res.strip())
        res = res.strip()
        if res:
            self.reply(res)

    def handle(self):
        """ handle an event. """
        if not self._loader:
            raise ELOADER
        self.parse()
        if not self._parsed.cmd:
            self.ready()
            return self
        self._loader.prep(self)
        self.dispatch()
        if self._batch:
            self.show()
        self.ready()
        return self

    def format(self, keys=[], skip=[], nokeys=False, reverse=False) -> str:
        """ return a displayable string from a list of attributes. """
        keys = keys or self.keys()
        if reverse:
            keys = reversed(list(keys))
        result = []
        for key in reversed(list(keys)):
            if key == "default":
                continue
            if key.startswith("_"):
                continue
            if key in skip:
                continue
            if nokeys:
                result.append("%-8s" % str(self[key]))
            else:
                result.append("%10s" % "%s=%s" % (key, str(self[key])))
        txt = " ".join(result)
        return txt.rstrip()

    def grep(self, val: str) -> 'Object':
        """ return a new Object with the values that match provided val argument. """
        o = Object()
        for key, value in self.items():
            if val in str(value) and value not in o:
                o[key] = value
        return o

    def isSet(self) -> bool:
        """ check whether ready flag is set. """
        return self._ready.isSet()

    def load(self, path: str, force=False, skip=[], full=True, run=False) -> 'Object':
        """ load a json file into this object. use skip as a list of keys to skip. """
        import json, logging
        if not cfg.workdir in path:
            path = os.path.join(cfg.workdir, path)
        logging.debug("load %s" % os.sep.join(path.split(os.sep)[-3:]))
        ondisk = self.read(path)
        self._container.path = path
        fromdisk = json.loads(ondisk)
        if "signature" in fromdisk:
            if not verify_signature(fromdisk["data"], fromdisk["signature"]) and not force:
                logging.warn("mismatch %s" % os.sep.join(path.split(os.sep)[-3:]))
        if "data" in fromdisk:
            self.update(slice(fromdisk["data"], skip=skip, full=full)) 
            self._container.update(slice(fromdisk, skip=["data"]))
        cache[path] = self
        if run:
            thr = self.run()
            thr.join()
        return self

    def loads(self, jsonstring: str):
        """ update with deconstructed (dict) json string. """
        import json
        self.update(json.loads(jsonstring))
        return self

    def merge(self, obj) -> 'Object':
        """ merge an object into this on, only set keys that are already present. """
        for k, v in obj.items():
            if v:
                self[k] = v
        return self

    def nice(self) -> str:
        """ return a nicyfied JSON dump. """
        import json
        return json.dumps(self, default=smooth, indent=4, sort_keys=True)

    def parse(self):
        """ parse the event. """
        self._parsed.parse(self.txt)

    def pure(self) -> str:
        """ return a sliced (no _ keys), indent=4, sort_key is True, json dump. """
        import json
        return json.dumps(slice(self, full=False), indent=4, sort_keys=True)

    def prepare(self) -> str:
        """ prepare the object and return a string containing the "data" part. """
        import json, time
        todisk = Object()
        todisk.data = dumped(slice(self, skip=["_container", "_parsed"], full=True))
        todisk.data.__typed__ = self.__class__.__name__
        if "prefix" not in todisk:
            todisk.prefix = get_prefix(self)
        todisk.saved = time.ctime(time.time())
        try:
            todisk.signature = make_signature(todisk["data"])
        except:
            pass
        try:
            result = json.dumps(todisk, default=smooth, indent=4, ensure_ascii=False, sort_keys=True)
        except TypeError:
            raise ENOJSON(todisk.prefix)
        return result

    def raw(self, txt):
        """ raw output. """
        print(txt)

    def read(self, path: str) -> str:
        """ read a json dump from given path, returning the json string with comments stripped. """
        try:
            f = open(path, "r", encoding="utf-8")
        except FileNotFoundError:
            return "{}"
        res = ""
        for line in f:
            if not line.strip().startswith("#"):
                res += line
        if not res.strip():
            return "{}"
        f.close()
        return res

    def ready(self) -> bool:
        """ signal this object as "ready". """
        return self._ready.set()

    def reply(self, txt):
        """ give a reply to the origin of this event. """
        if not self._batch:
             self.raw(txt)
        self.add(txt)

    def say(self, channel, txt, type="normal"):
        """ output to channel. """
        self.raw(txt)
        
    def show(self):
        """ show the event on the server is originated on. """
        for txt in self._result:
            self.raw(txt)
        
    def register(self, key, val, force=False) -> None:
        """ register key, value .. throw an exception if key is already set. """
        if key in self and not force:
            raise ESET(key)
        self[key] = val

    def save(self, path="", stime="") -> str:
        """ save a static (fix filepath) version of this object. """
        return self.sync(path=path, stime=stime)

    def search(self, name:str) -> 'Object':
        """ search this objects keys skipping keys that start with "_". """
        o = Object()
        for key, value in self.items():
            if key.startswith("_"):
                continue
            if key in name:
                o[key] = value
            elif name in key.split(".")[-1]:
                o[key] = value
        return o

    def sync(self, path="", stime="", workdir="") -> str:
        """ sync to disk using provided/created path. this sync() method does the real saving to disk. """
        import fcntl, logging, time
        if not path:
            path = get_path(self)
        if not path:
            wd = workdir or cfg.workdir
            path = os.path.join(wd, get_prefix(self))
            if stime:
                path = os.path.join(path, stime)
            else:
                path = os.path.join(path, rtime())
        if cfg.workdir not in path:
            p = path.split(os.sep)[-2:]
            got = os.sep.join(p)
            path = os.path.join(workdir or cfg.workdir, got)
        path = os.path.abspath(os.path.normpath(path))
        if cfg.workdir not in path:
            raise EBORDER(path)
        if not allowchar(path):
            raise EFILENAME(path)
        logging.info("sync %s" % os.sep.join(path.split(os.sep)[-3:]))
        self._container.path = path
        self._container.saved = time.ctime()
        d, fn = os.path.split(path)
        if not os.path.exists(d):
            cdir(d)
        todisk = self.prepare()
        rpath = path + "_tmp"
        try:
            datafile = open(rpath, 'w')
            fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except BlockingIOError:
            datafile = open(rpath, 'w')
            fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % ("%s characters" % len(todisk), time.ctime(time.time())))
        datafile.write(todisk)
        datafile.write("\n")
        datafile.flush()
        os.rename(rpath, path)
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        return path

    def upgrade(self, input: dict) -> 'Object':
        """ upgrade this object with a dictionary, only setting attributes if they do not exist. """
        for k, v in input.items():
            if k.startswith("_"):
                continue
            if k in self:
                if not self[k]:
                    self[k] = input[k]
        return self

    def wait(self, timeout=None) -> list:
        """ wait for this object's ready flag and join all thread that have been started in handling this object. """
        result = []
        self._ready.wait(timeout)
        for thr in self._thrs:
            try:
                res = thr.join(timeout)
                result.append(res)
                self._thrs.remove(thr)
            except RuntimeError:
                pass
        return result

class Default(Object):

    """ object with a "default" value. standard default return is an empty string. """

    default = ""

    def __init__(self, *args, **kwargs):
        Object.__init__(self, *args, **kwargs)
        self._default = kwargs.get("_default", str)

    def __getattr__(self, name):
        """ override Object.__getattr__. """
        try:
            return super().__getattr__(name)
        except (AttributeError, KeyError):
            if self.default != None:
                self[name] = self["_default"](self.default)
            else:
                self[name] = self["_default"]()
        return self[name]

class Register(Object):

    """ class to register key,value pairs. """

    def register(self, key, val, force=False) -> None:
        """ register value with key, using force argumentto determine to overwrite or not. """
        if key not in self:
            self[key] = []
        if not force and val in self.get(key, []):
            return
        self[key].append(val)

    def find(self, txt=None) -> type:
        """ search a register object for keys matching txt. """
        for key, value in self.items():
            if txt and txt in key:
                yield value
            else:
                yield value

class Config(Default):

    """ config class with a empty string as default. """

    def fromdisk(self, name, tpl={}) -> 'Object':
        """ load the config from file. """
        self.update(tpl)
        self.cache(os.path.join(cfg.workdir, "config", name))
        return self

class Parsed(Object):

    """ parsed contains all the arguments that are _parsed from an event. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = []
        self.cmd = ""
        self.command = ""
        self.counter = 0
        self.delta = 0
        self.disabled = []
        self.enabled = []
        self.fields = []
        self.index = None
        self.ignore = Register()
        self.notwant = Register()
        self.rest = ""
        self.start = None
        self.switch = Register()
        self.uniqs = Register()
        self.want = Register()
        self.words = []

    def parse(self, txt, cc=""):
        """ parse txt to determine cmnd, args, rest and other values. adds a _parsed object to the event. """
        txt = bytes(txt, "utf-8").decode("unicode_escape") 
        if cc or "cc" in self and self.cc:
            cc = cc or self.cc
            if txt[0] != cc:
                return
            txt = txt[1:]
        splitted = txt.split()
        quoted = False
        key2 = ""
        counter = -1
        for word in splitted:
            counter += 1
            if counter == 0:
                if self.command:
                    self.cmd = self.command
                    continue
                if word:
                    self.cmd = word.lower().strip()
                continue
            try:
                key, value = word.split("=", 1)
            except (IndexError, ValueError):
                key = ""
                value = word
            if "http" in key:
                key = ""
                value = word
            if value.startswith('"'):
                value = value[1:]
                key2 = key
                quoted = True
            if quoted and '"' in value:
                value = value.replace('"', '')
                quoted = False
                key = key2
            if quoted:
                self.words.append(value)
                continue
            else:
                self.words.append(value)
                value = " ".join(self.words)
                self.words = []
            if "http" in value:
                self.args.append(value)
                self.rest += value + " "
                continue
            if key == "index":
                self.index = int(value)
                continue
            if key == "start":
                self.start = parse_time(value)
                continue
            if key == "end":
                self.end = parse_time(value)
                continue
            if key and value:
                value = value.replace('"', '')
                post = value[0]
                last = value[-1]
                pre = key[0]
                op = key[-1]
                if key.startswith("!"):
                    key = key[1:]
                    self.switch.register(key, value)
                    continue
                if post == "-":
                    value = value[1:]
                    self.ignore.register(key, value)
                    continue
                if op == "-":
                    key = key[:-1]
                    self.notwant.register(key, value)
                    continue
                if pre == "^":
                    key = key[1:]
                    self.uniqs.register(key, value)
                if last == "-":
                    value = value[:-1]
                self.want.register(key, value)
                if last == "-":
                    continue
                if key not in self.args:
                    self.args.append(key)
                if counter > 1:
                    self.fields.append(key)
                self.rest += key + " "
            elif value:
                post = value[0]
                last = value[-1]
                if post == "^":
                    value = value[1:]
                    self.uniqs.register(value, "")
                if value.startswith("+") or value.startswith("-"):
                    v = value[1:]
                    try:
                        val = int(value)
                        self.delta = 0 + (val * 60*60)
                        if val >= -10 and val <= 10:
                            self.karma = val
                    except ValueError:
                        pass
                    if post == "-":
                        if v not in self.enabled:
                            self.enabled.append(v)
                            continue
                    elif post == "+":
                        if v not in self.disabled:
                            self.disabled.append(v)
                            continue
                if counter > 1:
                    self.fields.append(value)
                if len(splitted) == 2:
                    self.fields.append(value)                    
                self.args.append(value)

                self.rest += str(value) + " "
        self.rest = self.rest.strip()
        return self

## helpers

def allowchar(s) -> bool:
    """ check if s is a allowed filename string. """
    res = False
    for i in s:
        if i in allowedchars:
            res = True
        else:
            res = False
    return res

def cdir(path) -> str:
    """ create a directory. """
    import logging
    res = ""
    padje = ""
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        padje = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(padje)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
        except OSError as ex:
            logging.error(get_exception())
    return padje

def dumped(o) -> type:
    """ add type string to an object attributes. """
    t = type(o)
    if t == str:
        return o
    if t in [list, tuple]:
        l = []
        for obj in o:
            l.append(dumped(obj))
        return l
    if "items" not in dir(o):
        return o
    oo = t()
    for k, v in o.items():
        try:
            v["__typed__"] = str(type(v))
        except:
            pass
        oo[k] = dumped(v)
    return oo

def fromstring(txt) -> str:
    """ derive an object from the repr string. """
    if " at " in txt:
        txt = txt.split(" at ")[0]
    if " of " in txt:
        t1, t2 = txt.split(" of ")
        txt = t2.split(".")[-1] + "." + t1.split(".")[-1]
    if "," in txt:
        txt = txt.split(",")[0]
    txt = txt.replace("<function ", "")
    return txt.strip()

def get_prefix(obj) -> str:
    """ return the object's prefix. """
    return name(obj).split(".")[-1].lower()

def get_path(obj) -> str:
    """ return the path used to store the object's json dump. """
    try:
        return obj._container.path
    except:
        pass

def get_saved(obj) -> str:
    """ return the saved attribue of an object. """
    p = ""
    if "_container" in obj:
        p = getattr(obj._container, "saved", "")
        if p:
            return p

def make_signature(obj) -> str:
    """ make an signature based on the values of an object. """
    import hashlib, json
    data = json.dumps(obj, indent=4, ensure_ascii=True, sort_keys=True)
    return str(hashlib.sha1(bytes(str(data), "utf-8")).hexdigest())

def name(obj) -> str:
    """ return the name destilled from the repr of an object. """
    txt = repr(obj)
    return fromstring(txt)

def rtime() -> str:
    """ return a filestamp usable in a filename. """
    import datetime
    res =  str(datetime.datetime.now()).replace(" ", os.sep)
    return res

def slice(obj, keys=[], skip=[], full=True, skipempty=True) -> 'Object':
    """ return a slice of an object. """
    t = type(obj)
    o = t()
    if not keys:
        keys = obj.keys()
    for key in keys:
        if key in skip:
            continue
        if not full and key.startswith("_"):
            continue
        val = obj.get(key, None)
        if skipempty and not val:
            continue
        if val and "keys" in dir(val):
            o[key] = slice(val, skip=skip, full=full, skipempty=skipempty)
        else:
            o[key] = val
    return o

def smooth(obj) -> 'Object':
    """ return a JSON representation of an object. """
    if type(obj) not in basic_types:
        return repr(obj)
    return obj

def verify_signature(obj, signature) -> bool:
    """ verify an object's integrty based on the object's signature. """
    signature2 = make_signature(obj)
    return signature2 == signature

def ohelp(o) -> None:
    """ object help. """
    for key in dir(o):
        if key.startswith("_"):
            continue
        oo = getattr(o, key, None)
        if oo:
            try:
                print("%s.%s(%s)\n %s" % (oo.__module__, oo.__name__, oo.__base__.__name__, oo.__doc__))
            except:
                try:
                    print("%s.%s\n %s" % (oo.__module__, oo.__name__, oo.__doc__))
                except:
                    print("%s\n %s" % (key, oo.__doc__))

## vars

cfg = Config()
cfg.workdir = os.path.join(os.path.expanduser("~"), ".obj")
cfg.level = "error"

#:
cache = Object()

#:
nodict_types = [str, int, float, bool, None]

#:
basic_types = [dict, list, str, int, float, bool, None]

#:
allowedchars = string.ascii_letters + string.digits + '_+/$.-'

#:
headertxt = '''# this is an OBJ file, %s
#
# the program can edit this file !!
#
# saved on %s
'''
