#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# OBJ - timestamped JSON objects
#
# obj/cmds.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 24-02-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" commands module. """

## imports

from obj import _starttime, name
from obj import Object, __version__
from obj.core import db, fleet, launcher, loader
from obj.time import elapsed, to_day, day
from obj.trace import get_exception

import os
import time

## classes

class Email(Object):
    """ email class. """

class Entry(Object):
    """ class used to save entry data. """
    pass

class Log(Entry):
    """ a log entry to log what is happening on the day. """
    pass

class Rss(Entry):
    """ rss entry that consists of the url of the feed. """
    pass

class Shop(Entry):
    """ shopping item for the wife to shop. """
    pass

class Todo(Entry):
    """ todo entry """
    pass

class Tomorrow(Entry):
    """ todo for tomorrow to be done. """
    pass

class Watch(Entry):
    """ files to watch. """
    pass

## cmds

psformat = "%-4s %-8s %-60s"

def attr(event):
    """ show attributes of an object type. """
    from obj import cfg
    if not event._parsed.args:
         options = ",".join(os.listdir(cfg.workdir))
         event.reply("attr <prefix>, use one of %s" % options)
         return
    prefix = event._parsed.args[0]
    try:
        keys = event._parsed.args[1:]
    except:
        keys = []
    obj = db.last(prefix)
    if obj:
        for key in keys:
            obj = getattr(obj, key, None)
        try:
            res = obj.keys()
        except AttributeError:
            res = []
            for o in obj:
                res.extend(o.keys())
            res = set(res)
        event.reply("\n".join(sorted([x for x in res if not x.startswith("_")])))
    else:
        event.reply("no %s prefix found." % event._parsed.args[0])

def cfg(event):
    """ show config. """
    from obj import cfg
    event.reply(cfg)

def cmds(event):
    """ show commands registered. """
    cli = fleet[0]
    event.reply(",".join(cli._handlers.keys()))

def deleted(event):
    """ show deleted records. """
    event.nodel = True
    nr = 0
    for obj in db.selected(event):
        if "deleted" in dir(obj) and obj.deleted:
            nr += 1
            event.display(obj, str(nr))
    
def ed(event):
    """ object editor. """
    args = event._parsed.args
    if not args:
        event.reply("ed <name> <key> <value>")
        return
    n = args[0]
    obj = None
    for thr in sorted(lhr.running(), key=lambda x: name(x)):
        if n.upper() not in str(thr).upper():
            continue
        try:
            obj = thr._func.__self__
            id = obj.id
        except:
            continue 
        if not obj:
            continue
        if n.lower() in id.lower():
            break
    if len(args) == 1:
        print(obj)
        return 
    key = args[1]
    if key not in obj:
        event.reply("EKEY %s" % key)
        return
    if len(args) == 2:
        event.reply('%s="%s"' % (key, obj[key]))
        return
    val = ""
    if len(args) > 3:
        val = args[2:]
    if len(args) == 3:
        val = args[2]
    try:
        val = int(val)
    except:
        try:
            val = float(val)
        except:
            pass
    if type(val) == list:
        obj[key] = list(val)
    elif val in ["True", "true"]:
        obj[key] = True
    elif val in ["False", "false"]:
        obj[key] = False
    else:
        try:
            val = '"%s"' % val
            obj[key] = json.loads(val)
        except (SyntaxError, ValueError) as ex:
            event.reply("%s: %s" % (str(ex), val))
            return

def find(event):
    """ present a list of objects based on prompt input. """
    nr = 0
    seen = []
    for obj in db.selected(event):
        for key in event._parsed.uniqs.keys():
            val =  getattr(obj, key, None)
            if val and val in seen:
                continue
            else:
                seen.append(val)
        if "d" in event._parsed.enabled:
            event.reply(obj)
        else:
            event.display(obj, txt=str(nr))
        nr += 1

def first(event):
    """ show the first record matching the given criteria. """
    obj = db.first(*event._parsed.args)
    if obj:
        event.reply(obj)

def fix(event):
    """ fix a object by loading and saving it. """
    fn = event._parsed.rest
    fn = os.path.abspath(fn)
    if not fn:
        event.reply("fix <path>")
        return
    if not os.path.isfile(fn):
        event.reply("%s is not a file" % fn)
        return
    o = Object()
    o.load(fn)
    p = o.save()
    event.reply("saved %s" % p)

def kernel(event):
    """ display generic loader in obj.core. """
    event.reply(loader)

def last(event):
    """ show last objectect matching the criteria. """
    if not event._parsed.args:
        event.reply("last <prefix> <value>")
        return
    obj = db.last(*event._parsed.args)
    if obj:
        event.reply(obj)

def log(event):
    """ log some text. """
    if not event._parsed.rest:
        event.reply("log <item>")
        return
    o = Log()
    o.log = event._parsed.rest
    o.save()
    event.reply("ok 1")

def ls(event):
    """ show subdirs in working directory. """
    from obj import cfg
    prefixes = os.listdir(cfg.workdir)
    if prefixes:
        event.reply(" ".join(prefixes))

def mbox(event):
    """ convert emails to objects. """
    if not event._parsed.rest:
        event.reply("mbox <path>")
        return
    fn = os.path.expanduser(event._parsed.args[0])
    event.reply("reading from %s" % fn)
    nr = 0
    if os.path.isdir(fn):
        thing = mailbox.Maildir(fn, create=False)
    elif os.path.isfile(fn):
        thing = mailbox.mbox(fn, create=False)
    else:
        event.reply("need a mbox or maildir.")
        return
    try:
        thing.lock()
    except FileNotFoundError:
        pass
    for m in thing:
        try:
            o = Email()
            o.update(m.items())
            try:
                sdate = os.sep.join(to_date(o.Date).split())
            except AttributeError:
                sdate = None
            o.text = ""
            for payload in m.walk():
                if payload.get_content_type() == 'text/plain':
                    o.text += payload.get_payload()
            o.text = o.text.replace("\\n", "\n")
            if sdate:
                o.save(stime=sdate)
            else:
                o.save()
            nr += 1
        except:
            logging.error(get_exception())
    if nr:
        event.reply("ok %s" % nr)

def ps(event):
    """ show running threads. """
    up = elapsed(int(time.time() - _starttime))
    result = []
    for thr in sorted(launcher.running(), key=lambda x: name(x)):
        obj = Object(vars(thr))
        if "sleep" in obj:
            up = obj.sleep - int(time.time() - obj._time.latest)
        else:
            up = int(time.time() - _starttime)
        thrname = name(thr)[1:] + ")"
        result.append((up, thrname, obj))
    nr = 0
    for up, thrname, obj in sorted(result, key=lambda x: x[0]):
        nr += 1
        res = psformat % (nr, elapsed(up), thrname[:30])
        event.reply(res.rstrip())

def reload(event):
    """ reload a plugin. """
    if not event._parsed.rest:
        event.reply(",".join([x.split(".")[-1] for x in loader.modules("obj")]))
        return
    for modname in loader.modules("obj"):
        if event._parsed.rest not in modname:
            continue
        try:
            mod = loader.reload(modname, False)
            if mod:
                event.reply("ok %s" % modname)
            else:
                event.reply("no %s module foudn to reload." % modname)
        except (AttributeError, KeyError) as ex:
            event.reply("%s %s" % (modname, str(ex)))

def restore(event):
    """ set deleted=False in selected records. """
    nr = 0
    event.nodel = True
    for obj in db.selected(event):
        obj.deleted = False
        obj.sync()
        nr += 1
    event.reply("ok %s" % nr)

def rm(event):
    """ set deleted flag on objects. """
    nr = 0
    for obj in db.selected(event):
        obj.deleted = True
        obj.sync()
        nr += 1
    event.reply("ok %s" % nr)

def shop(event):
    """ add a shopitem to the shopping list. """
    if not event._parsed.rest:
        event.reply("shop <item>")
        return
    o = Shop()
    o.shop = event._parsed.rest
    o.save()
    event.reply("ok 1")

def start(event):
    """ start a plugin. """
    if not event._parsed.args:
        modnames = [x.split(".")[-1] for x in loader.modules("bot")]
        res = [x.split(".")[-1] for x in modnames]
        event.reply("choose one of %s" % ",".join(sorted(res)))
        return
    n = event._parsed.args[0]
    mod = loader.direct(n)
    try:
        mod.start()
        event.reply("ok %s" % n)
    except:
        event.reply(get_exception())

def stop(event):
    """ stop a plugin. """
    from obj.core import launcher
    if not event._parsed.rest:
        event.reply("stop what ?")
        return
    n  = event._parsed.args[0]
    if n:
        launcher.kill(n.upper())

def test(event):
    """ say hello. """
    event.reply("hello %s" % event.origin)

def tests(event):
    """ run all possble commands. """
    cli = fleet[0]
    for cmd in loader._handlers.keys():
        if cmd == "tests":
            continue
        e = cli.event(cmd)
        e.handle()

def today(event):
    """ show last week's logged objects. """
    event._parsed.start = to_day(day())
    event._parsed.end = time.time()
    nr = 0
    for obj in db.selected(event):
        nr += 1
        event.display(obj, str(nr))

def todo(event):
    """ log a todo item. """
    if not event._parsed.rest:
        event.reply("todo <item>")
        return
    o = Todo()
    o.todo = event._parsed.rest
    o.save()
    event.reply("ok 1")

def tomorrow(event):
    """ show todo items for tomorrow. """
    if not event._parsed.rest:
        event.reply("tomorrow <item>")
        return
    o = Tomorrow()
    o.tomorrow = event.txt.replace("tomorrow", "").strip()
    o.save()
    event.rpely("ok 1")

def version(event):
    """ show program's version. """
    event.reply("OBJ %s" % __version__)

def week(event):
    """ show last week's logged objects. """
    event._parsed.start = to_day(day()) - 7 * 24 * 60 * 60
    event._parsed.end = time.time()
    nr = 0
    for obj in db.selected(event):
        nr += 1
        event.display(obj, str(nr))

def yesterday(event):
    """ show last week's logged objects. """
    event._parsed.start = to_day(day()) - 24 * 60 * 60
    event._parsed.end = to_day(day())
    nr = 0
    for obj in db.selected(event):
        nr += 1
        event.display(obj, str(nr))
